import aiohttp

from .device import Device

class Sensor(Device):
    '''
    Represents Sensor SmartRent device
    '''
    def __init__(
        self,
        email: str,
        password: str,
        device_id: str,
        aiohttp_session:aiohttp.ClientSession=None
    ):
        super().__init__(email, password, device_id, aiohttp_session)
        self._locked = None
        self._notification = None


from typing import Optional
import logging

from .device import Device
from .utils import Client

_LOGGER = logging.getLogger(__name__)

class DoorLock(Device):
    '''
    Represents Lock SmartRent device
    '''
    def __init__(
        self,
        device_id: int,
        client: Client
    ):
        super().__init__(device_id, client)
        self._locked = None
        self._notification = None


    def get_notification(self) -> Optional[str]:
        '''
        Notification message for lock
        '''
        return self._notification


    def get_locked(self) -> Optional[bool]:
        '''
        Gets state from lock
        '''
        return self._locked


    async def async_set_locked(self, value: bool):
        '''
        Sets state for lock
        '''

        self._locked = value

        # Convert to lowercase just like SmartRent website does
        value = str(value).lower()

        await self._async_send_command(
            attribute_name='locked',
            value=value
        )


    def _fetch_state_helper(self, data: dict):
        '''
        Called when ``_async_fetch_state`` returns info

        ``data`` is dict of info passed in by ``_async_fetch_state``
        '''
        self._name = data['name']

        attrs = self._structure_attrs(data['attributes'])

        self._locked = bool(attrs['locked'] == 'true')
        self._notification = attrs['notifications']


    def _update_parser(self, event: dict):
        '''
        Called when ``_async_update_state`` returns info

        ``event`` dict passed in from ``_async_update_state``
        '''
        _LOGGER.info('Updating DoorLock')

        if event.get('name') == 'locked':
            self._locked = bool(event['last_read_state'] == 'true')

        if event.get('name') == 'notifications':
            self._notification = event.get('last_read_state')



import logging
from typing import List, Union

import aiohttp

from .lock import DoorLock
from .thermostat import Thermostat
from .utils import Client

_LOGGER = logging.getLogger(__name__)


class API():
    def __init__(self, email: str, password: str, aiohttp_session: aiohttp.ClientSession = None):
        '''
        Represents overall SmartRent api

        ``email`` is the email address for your SmartRent account

        ``password`` you know what it is

        ``aiohttp_session`` (optional) uses the aiohttp_session that is passed in
        '''
        self._device_list = []
        self._email = email
        self._password = password
        self._session = aiohttp_session

        self.client: Client = Client(email, password, aiohttp_session)


    async def async_fetch_devices(self):
        '''
        Fetches list of devices by calling SmartRent api
        '''
        _LOGGER.info('Fetching devices via API...')
        await self.client.async_refresh_token()
        data = await self.client.async_get_devices_data()
        _LOGGER.info('Got devices!')

        for device in data:
            device_id = device.get('id')
            device_type = device.get('type')

            device_object:Union[Thermostat, DoorLock] = None

            if device_type == 'thermostat':
                device_object = Thermostat(device_id, self.client)

            elif device_type == 'entry_control':
                device_object = DoorLock(device_id, self.client)

            if device_object:
                # pass in intial device config
                device_object._fetch_state_helper(device)

                # add device to device_list
                self._device_list.append(device_object)


    def get_locks(self) -> List[DoorLock]:
        '''
        Gets list of DoorLocks
        '''
        return [x for x in self._device_list if isinstance(x, DoorLock)]


    def get_thermostats(self) -> List[Thermostat]:
        '''
        Gets list of Thermostats
        '''
        return [x for x in self._device_list if isinstance(x, Thermostat)]


async def async_login(
    email: str,
    password: str,
    aiohttp_session: aiohttp.ClientSession = None
) -> API:
    '''
    Logs into SmartRent and retruns an `API` object

    ``email`` is the email address for your SmartRent account

    ``password`` you know what it is

    ``aiohttp_session`` (optional) uses the aiohttp_session that is passed in
    '''

    smart_rent_api = API(email, password, aiohttp_session)
    await smart_rent_api.async_fetch_devices()

    return smart_rent_api



from typing import Optional, Union
import logging

from .device import Device
from .utils import Client

_LOGGER = logging.getLogger(__name__)

class Thermostat(Device):
    '''
    Represents Thermostat SmartRent device
    '''
    def __init__(
        self,
        device_id: int,
        client: Client
    ):
        super().__init__(device_id, client)
        self._mode = None
        self._fan_mode = None
        self._cooling_setpoint = None
        self._heating_setpoint = None
        self._current_humidity = None
        self._current_temp = None


    def get_mode(self) -> Optional[str]:
        '''
        Gets mode from thermostat
        '''
        return self._mode


    def get_fan_mode(self) -> Optional[str]:
        '''
        Gets fan mode from thermostat
        '''
        return self._fan_mode


    def get_cooling_setpoint(self) -> Optional[int]:
        '''
        Gets cooling setpoint from thermostat
        '''
        return self._cooling_setpoint


    def get_heating_setpoint(self) -> Optional[int]:
        '''
        Gets heating setpoint from thermostat
        '''
        return self._heating_setpoint


    def get_current_humidity(self) -> Optional[int]:
        '''
        Gets current humidity from thermostat
        '''
        return self._current_humidity


    def get_current_temp(self) -> Optional[int]:
        '''
        Gets current temperature from thermostat
        '''
        return self._current_temp


    async def async_set_heating_setpoint(self, value: int):
        '''
        Sets heating setpoint

        ``value`` str or int representing temperature to set
        '''
        await self._async_send_command(
            attribute_name='heating_setpoint',
            value=value
        )

        self._heating_setpoint = int(value)


    async def async_set_cooling_setpoint(self, value: int):
        '''
        Sets cooling setpoint

        ``value`` str or int representing temperature to set
        '''
        await self._async_send_command(
            attribute_name='cooling_setpoint',
            value=value
        )

        self._cooling_setpoint = int(value)


    async def async_set_mode(self, mode: str):
        '''
        Sets thermostat mode

        ``mode`` str. One of ``['aux_heat', 'heat', 'cool', 'auto', 'off']``
        '''
        if mode not in ['aux_heat', 'heat', 'cool', 'auto', 'off']:
            return

        await self._async_send_command(
            attribute_name='mode',
            value=mode
        )

        self._mode = mode


    async def async_set_fan_mode(self, fan_mode: str):
        '''
        Sets thermostat fan mode

        ``value`` str. One of ``['auto', 'on']``
        '''
        if fan_mode not in ['on', 'auto']:
            return

        await self._async_send_command(
            attribute_name='fan_mode',
            value=fan_mode,
        )

        self._fan_mode = fan_mode


    def _fetch_state_helper(self, data: dict):
        '''
        Called when ``_async_fetch_state`` returns info

        ``data`` is dict of info passed in by ``_async_fetch_state``
        '''
        self._name = data['name']

        attrs = self._structure_attrs(data['attributes'])

        float_to_int = lambda x: int(float(x))
        self._current_temp = float_to_int(attrs['current_temp'])
        self._current_humidity = float_to_int(attrs['current_humidity'])

        self._cooling_setpoint = float_to_int(attrs['cooling_setpoint'])
        self._heating_setpoint = float_to_int(attrs['heating_setpoint'])

        self._mode = attrs['mode']
        self._fan_mode = attrs.get('fan_mode')


    def _update_parser(self, event: dict) -> None:
        '''
        Called when ``_async_update_state`` returns info

        ``event`` dict passed in from ``_async_update_state``
        '''
        _LOGGER.info('Updating Thermostat')
        if event.get('name') == 'current_humidity':
            self._current_humidity = int(event.get('last_read_state'))

        if event.get('name') == 'current_temp':
            self._current_temp = int(event.get('last_read_state'))

        if event.get('name') == 'heating_setpoint':
            self._heating_setpoint = int(event.get('last_read_state'))

        if event.get('name') == 'cooling_setpoint':
            self._cooling_setpoint = int(event.get('last_read_state'))

        if event.get('name') == 'mode':
            self._mode = event.get('last_read_state')

        if event.get('name') == 'fan_mode':
            self._fan_mode = event.get('last_read_state')



import asyncio
import logging

import aiohttp

_LOGGER = logging.getLogger(__name__)

SMARTRENT_BASE_URI     = 'https://control.smartrent.com/api/v2/'
SMARTRENT_SESSIONS_URI = SMARTRENT_BASE_URI + 'sessions'
SMARTRENT_HUBS_URI     = SMARTRENT_BASE_URI + 'hubs'
SMARTRENT_HUBS_ID_URI  = SMARTRENT_BASE_URI + 'hubs/{}/devices'


class SmartRentError(Exception):
    '''
    Base error for SmartRent
    '''


class InvalidAuthError(SmartRentError):
    '''
    Error related to invalid auth
    '''


class Client():
    def __init__(
        self,
        email: str,
        password: str,
        aiohttp_session: aiohttp.ClientSession
    ):
        '''
        Represents Cleint for SmartRent api.
        Usually shared between multiple devices for best performance.

        ``email`` is the email address for your SmartRent account

        ``password`` you know what it is

        ``aiohttp_session`` (optional) uses the aiohttp_session that is passed in
        '''
        self._email = email
        self._password = password

        self._im_session_owner = not bool(aiohttp_session)
        self._aiohttp_session = aiohttp_session if aiohttp_session else aiohttp.ClientSession()
        self.token = None


    def __del__(self):
        '''
        Handles delete of aiohttp session if class is tasked with it
        '''
        if not self._aiohttp_session.closed and self._im_session_owner:
            _LOGGER.info('%s: closing aiohttp session %s', str(self), self._aiohttp_session)
            asyncio.run(self._aiohttp_session.close())


    async def async_get_devices_data(self) -> dict:
        '''
        Gets device dictonary from SmartRent's api.
        Also handles retry if token is bad
        '''
        res = await self._async_get_devices_data()
        if not res:
            _LOGGER.warning('No devices returned. Trying again with updated token...')
            await self.async_refresh_token()

            res = await self._async_get_devices_data()
        return res


    async def _async_get_devices_data(self) -> dict:
        '''
        Gets device dictonary from SmartRent's api
        '''

        hubs_resp = await self._aiohttp_session.get(
            SMARTRENT_HUBS_URI,
            headers = {
                'authorization': f'Bearer {self.token}'
            }
        )
        hubs = await hubs_resp.json()

        devices_list = []
        for hub in hubs:
            devices_resp = await self._aiohttp_session.get(
                SMARTRENT_HUBS_ID_URI.format(hub['id']),
                headers = {
                    'authorization': f'Bearer {self.token}'
                }
            )
            devices = await devices_resp.json()

            for device in devices:
                _LOGGER.info('Found %s: %s', device['id'], device['name'])
                devices_list.append(device)

        return devices_list


    async def async_refresh_token(self) -> None:
        '''
        Refreshes API token from SmartRents
        '''
        result = await self._aiohttp_session.post(
            SMARTRENT_SESSIONS_URI,
            json={
                'email': self._email,
                'password': self._password
            }
        )

        result = await result.json()
        if not result.get('errors'):
            self.token = result['access_token']
        else:
            raise Exception(
                'Token not retrieved! '
                f'Loggin probably not successful: {result["errors"]}'
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['smartrent']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0', 'websockets>=10.1,<11.0']

setup_kwargs = {
    'name': 'smartrent.py',
    'version': '0.1.3',
    'description': 'Unofficial Python API for SmartRent devices',
    'long_description': "# SmartRent API\n\n[![PyPI version](https://img.shields.io/pypi/v/smartrent-py.svg?logo=pypi&logoColor=FFE873)](https://pypi.org/project/smartrent-py/)\n[![Supported Python versions](https://img.shields.io/pypi/pyversions/smartrent-py.svg?logo=python&logoColor=FFE873)](https://pypi.org/project/smartrent-py/)\n[![Downloads](https://static.pepy.tech/personalized-badge/smartrent-py?period=total&units=international_system&left_color=grey&right_color=orange&left_text=total%20downloads)](https://pepy.tech/project/smartrent-py)\n[![PyPI downloads](https://img.shields.io/pypi/dm/smartrent-py.svg)](https://pypistats.org/packages/smartrent-py)\n[![GitHub](https://img.shields.io/github/license/ZacheryThomas/smartrent.py.svg)](LICENSE.txt)\n\nUnofficial api for SmartRent devices\n\nUses websockets for communication and supports async functions\n\n## Known Devices supported\n### Locks\n* Yale YRD256\n\n### Thermostats\n* Honeywell T6 Pro (TH6320ZW2003)\n\n\n# Usage\n\n## Installing\n\n```bash\npip install smartrent.py\n```\n\n## Getting an API Object\nIn order to get an api object to interact with, you must login with the `async_login` function. This starts and handles a web session with SmartRent's webserver.\n\n```python\nimport asyncio\n\nfrom smartrent import async_login\n\nasync def main():\n    api = await async_login('<EMAIL>', '<PASSWORD>')\n\nasyncio.run(main())\n```\n\n## Getting Devices\nYou can get lists of your devices from the api with the `get_locks` and `get_thermostats` functions. You can then interact with the devices with thier getter and setter functions.\n\n```python\nimport asyncio\n\nfrom smartrent import async_login\n\nasync def main():\n    api = await async_login('<EMAIL>', '<PASSWORD>')\n\n    lock = api.get_locks()[0]\n    locked = lock.get_locked()\n\n    if not locked:\n        await lock.async_set_locked(True)\n\nasyncio.run(main())\n```\n\n## Automatic Updating\nIf you need to get live updates to your device object from SmartRent, you can do that by calling `start_updater`. You can stop getting updates by calling `stop_updater`.\n\nYou can also set a callback function via `set_update_callback` that will be called when an update is triggered.\n\nFor example, if you want to set your thermostat to `Dad Mode` you can trigger an event every time the `cooling_setpoint` is changed and just change it back to your own desired value.\n```python\nimport asyncio\n\nfrom smartrent import async_login\n\nasync def main():\n    api = await async_login('<EMAIL>', '<PASSWORD>')\n\n    thermo = api.get_thermostats()[0]\n    thermo.start_updater()\n\n    CONSTANT_COOL = 80\n\n    async def on_evt():\n        if CONSTANT_COOL != thermo.get_cooling_setpoint():\n            await thermo.async_set_cooling_setpoint(CONSTANT_COOL)\n\n    thermo.set_update_callback(on_evt)\n\n    while True:\n        await asyncio.sleep(60)\n\nasyncio.run(main())\n```\n",
    'author': 'Zachery Thomas',
    'author_email': 'zacherythomas12@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zacherythomas/smartrent.py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1',
}


setup(**setup_kwargs)



import json
import asyncio
import inspect

from typing import Union
import logging

import websockets
from .utils import Client

_LOGGER = logging.getLogger(__name__)

SMARTRENT_WEBSOCKET_URI   = 'wss://control.smartrent.com/socket/websocket?token={}&vsn=2.0.0'
JOINER_PAYLOAD            = '["null", "null", "devices:{device_id}", "phx_join", {{}}]'
COMMAND_PAYLOAD           = '["null", "null", "devices:{device_id}", "update_attributes", ' \
                            '{{"device_id": {device_id}, ' \
                            '"attributes": [{{"name": "{attribute_name}", "value": "{value}"}}]}}]'

class Device():
    '''
    Base class for SmartRent devices
    '''
    def __init__(
        self,
        device_id:Union[str, int],
        client:Client
    ):
        self._device_id = int(device_id)
        self._name: str = ''
        self._update_callback_func = None
        self._updater_task = None

        self._client: Client = client


    @staticmethod
    def _structure_attrs(attrs: list):
        '''
        Converts device json object to hirearchical list of attributes

        ``attrs``: List of device attributes
        '''
        structure = {}

        for attr in attrs:
            name = attr.get('name')
            state = attr.get('state')

            structure[name] = state
        return structure


    def _fetch_state_helper(self, data: dict):
        '''
        Called by ``_async_fetch_state``

        Converts event dict to device param info
        '''
        raise NotImplementedError


    async def _async_fetch_state(self):
        '''
        Fetches device information from SmartRent api

        Calls ``_fetch_state_helper`` so device can parse out info and update its state.

        Calls function passed into ``set_update_callback`` if it exists.
        '''
        _LOGGER.info('%s: Fetching Status res page call...', self._name)
        data = await self._client.async_get_devices_data()

        # Find device id that belongs to me then call _fetch_state_helper
        for device in data:
            if device['id'] == self._device_id:
                self._fetch_state_helper(device)

                if self._update_callback_func:
                    if inspect.iscoroutinefunction(self._update_callback_func):
                        await self._update_callback_func()
                    else:
                        self._update_callback_func()


    def start_updater(self):
        '''
        Starts running ``update_state`` in the background
        '''
        if not self._updater_task:
            _LOGGER.info('%s: Starting updater task', self._name)
            self._updater_task = asyncio.create_task(self._async_update_state())
        else:
            _LOGGER.warning('%s: Updater task already running', self._name)


    def stop_updater(self):
        '''
        Stops running ``update_state`` in the background
        '''
        if self._updater_task:
            _LOGGER.info('%s: Stopping updater task', self._name)
            self._updater_task.cancel()


    def set_update_callback(self, func) -> None:
        '''
        Allows callback to be fired when ``_async_update_state``
        or ``_async_fetch_state`` gets new information
        '''
        self._update_callback_func = func


    def _update_parser(self, event: dict) -> None:
        '''
        Called by ``_async_update_state``

        Converts event dict to device param info
        '''
        raise NotImplementedError


    async def _async_update_state(self):
        '''
        Connects to SmartRent websocket and listens for updates.
        To be ran in the background. You can call ``start_updater`` and ``stop_updater``
        to turn ``_async_update_state`` on or off.

        Calls ``_update_parser`` method for device when event is found
        '''
        uri = SMARTRENT_WEBSOCKET_URI.format(self._client.token)

        async with websockets.connect(uri) as websocket:
            joiner = JOINER_PAYLOAD.format(device_id=self._device_id)
            await websocket.send(joiner)

            while True:
                try:
                    resp = await websocket.recv()

                    formatted_resp = json.loads(f'{{"data":{resp}}}')['data'][4]

                    if formatted_resp.get('type'):
                        event = (
                            f'{formatted_resp.get("type", ""):<15} -> '
                            f'{formatted_resp.get("name", ""):<15} -> '
                            f'{formatted_resp.get("last_read_state", ""):<20}'
                        )
                        _LOGGER.info('%s %s',self._name, event)
                    else:
                        _LOGGER.info('%s %s',self._name, resp)

                    self._update_parser(formatted_resp)

                    if self._update_callback_func:
                        if inspect.iscoroutinefunction(self._update_callback_func):
                            await self._update_callback_func()
                        else:
                            self._update_callback_func()

                except (
                    websockets.exceptions.ConnectionClosedError,
                    websockets.exceptions.ConnectionClosedOK
                ) as exc:
                    _LOGGER.warning('%s: Got excpetion: %s', self._name, exc)

                    _LOGGER.info('%s: Getting new token', self._name)
                    await self._client.async_refresh_token()

                    # Lets fetch device state just to make sure
                    # we didn't miss anything wile the socket was down
                    _LOGGER.info('%s: Fetching current device status...', self._name)
                    await self._async_fetch_state()

                    _LOGGER.info('%s: Reconnecting to Websocket...', self._name)
                    uri = SMARTRENT_WEBSOCKET_URI.format(self._client.token)

                    websocket = await websockets.connect(uri)
                    _LOGGER.info('%s: Connected!', self._name)

                    _LOGGER.info(
                        '%s: Joining topic for %s:%s ...',
                        self._name,
                        self._name,
                        self._device_id
                    )
                    joiner = JOINER_PAYLOAD.format(device_id=self._device_id)
                    await websocket.send(joiner)


    async def _async_send_command(self, attribute_name: str, value: str):
        '''
        Sends command to SmartRent websocket

        ``attribute_name`` string of attribute to change
        ``value`` value for that attribute to be changed to
        '''
        payload = COMMAND_PAYLOAD.format(
            attribute_name=attribute_name,
            value=value,
            device_id=self._device_id
        )

        await self._async_send_payload(payload)


    async def _async_send_payload(self, payload: str):
        '''
        Sends payload to SmartRent websocket

        ``payload`` string of device attributes
        '''
        _LOGGER.info('sending payload %s', payload)

        async def sender(uri: str, payload: str):
            async with websockets.connect(uri) as websocket:
                joiner = JOINER_PAYLOAD.format(device_id=self._device_id)

                # Join topic given device id
                await websocket.send(joiner)
                # Send payload
                await websocket.send(payload)

        try:
            uri = SMARTRENT_WEBSOCKET_URI.format(self._client.token)

            await sender(uri, payload)

        except websockets.exceptions.InvalidStatusCode as exc:
            _LOGGER.debug('Possible issue during send_payload: %s', exc)

            # update token once
            await self._client.async_refresh_token()

            uri = SMARTRENT_WEBSOCKET_URI.format(self._client.token)

            await sender(uri, payload)
